/*
 * token.c --
 *
 *	Lifting the tokenizer into the Tcl level.
 *
 * Copyright (C) 1999 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: token.c,v 1.4 1999/11/20 17:04:06 aku Exp $
 */


#include "ptoolsInt.h"

/*
 * Forward declarations of internal procedures.
 */

static int
TokenizeWords _ANSI_ARGS_ ((Tcl_Interp* interp, char* script, int len,
			    int nested, PTTokenList* tlist));

static int
PreorderTraverse _ANSI_ARGS_ ((Tcl_Interp*   interp,
			       Tcl_Token*    tokenPtr,
			       int           numTokens,
			       PT_TokenProc* callback,
			       ClientData    clientData,
			       int           level));
static int
PostorderTraverse _ANSI_ARGS_ ((Tcl_Interp*   interp,
				Tcl_Token*    tokenPtr,
				int           numTokens,
				PT_TokenProc* callback,
				ClientData    clientData,
				int           level));

static int
PlainList _ANSI_ARGS_ ((Tcl_Interp*   interp,
			ClientData clientData, int level,
			Tcl_Token* tokenPtr));

static int
CallScript _ANSI_ARGS_ ((Tcl_Interp*   interp,
			 ClientData clientData, int level,
			 Tcl_Token* tokenPtr));

static Tcl_Obj*
TokenTypeAsString _ANSI_ARGS_ ((Tcl_Token* t));

/*
 *------------------------------------------------------*
 *
 *	PTInitTokenList --
 *
 *	------------------------------------------------*
 *	Initializes the passed to token list.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Changes the passed token list structure.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

void
PTInitTokenList (t)
PTTokenList* t;
{
  t->numTokens = 0;
  t->tokenPtr  = (Tcl_Token*) NULL;
}

/*
 *------------------------------------------------------*
 *
 *	PTClearTokenList --
 *
 *	------------------------------------------------*
 *	Clears the token list passed into it.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Frees memory.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

void
PTClearTokenList (t)
PTTokenList* t;
{
  if (t->tokenPtr != (Tcl_Token*) NULL) {
    Tcl_Free ((char*) t->tokenPtr);
  }

  PTInitTokenList (t);
}

/*
 *------------------------------------------------------*
 *
 *	PTAddTokens --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'bc::token' command.
 *	See the manpages for details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

void
PTAddTokens (t, token, numTokens)
PTTokenList* t;
Tcl_Token*   token;
int          numTokens;
{
  t->tokenPtr   = (Tcl_Token*) Tcl_Realloc ((char*) t->tokenPtr,
					    (t->numTokens + numTokens)*
					    sizeof(Tcl_Token));
  t->numTokens += numTokens;

  memcpy (t->tokenPtr + (t->numTokens - numTokens),
	  token, numTokens * sizeof(Tcl_Token));
}

/*
 *------------------------------------------------------*
 *
 *	TokenizeWords --
 *
 *	------------------------------------------------*
 *	Creates a list of tokens for a script.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory
 *
 *	Result:
 *		A standard tcl return code.
 *
 *------------------------------------------------------*
 */

static int
TokenizeWords (interp, script, len, nested, tlist)
Tcl_Interp*  interp; /* Interpreter for error reporting */
char*        script;
int          len;
int          nested;
PTTokenList* tlist;
{
  /* Local state of the parsing */

  Tcl_Parse parse;
  Tcl_Token c;
  char *p, *next;
  int   bytesLeft, gotParse;

  PTInitTokenList (tlist);

  /* Parse the script one command at a time (each iteration of the loop gets
   * one) and add the resulting tokens to the returned list.
   */

  p         = script;
  bytesLeft = len;
  gotParse  = 0;

  while (bytesLeft > 0) {
    if (Tcl_ParseCommand (interp, p, bytesLeft, nested, &parse) != TCL_OK) {
      goto error;
    }

    gotParse = 1;

    /* Convert the known tokens of the current command into a structured list.
     */

    if (parse.commentSize > 0) {
      /* Handle a possible preceding comment.
       */

      c.type          = PT_TOKEN_COMMENT;
      c.start         = parse.commentStart;
      c.size          = parse.commentSize;
      c.numComponents = 0;

      PTAddTokens (tlist, &c, 1);
    }

    if (parse.numWords > 0) {
      /* Insert a virtual command token before the main array, then
       * add the found list as its subordinate.
       */

      c.type          = TCL_TOKEN_COMMAND;
      c.start         = parse.tokenPtr [0].start;
      c.size          = parse.tokenPtr [0].size;
      c.numComponents = parse.numTokens;

      PTAddTokens (tlist, &c, 1);
      PTAddTokens (tlist, parse.tokenPtr, parse.numTokens);
    }

    /*
     * Prepare local state for next command / iteration
     */

    next       = parse.commandStart + parse.commandSize;
    bytesLeft -= (next - p);
    p          = next;

    Tcl_FreeParse(&parse);
    gotParse = 0;
  }

  /* tlist is result! */

  return TCL_OK;

error:
  /*
   * Generate various pieces of error information, such as the line
   * number where the error occurred and information to add to the
   * errorInfo variable. Then free resources that had been allocated
   * to the command.
   */

  if (gotParse) {
#if 0
    /* Not yet, would have to copy LogCompilationInfo from the
     * tcl core.
     */

    commandLength = parse.commandSize;
    prev = '\0';
    if (commandLength > 0) {
      prev = parse.commandStart [commandLength-1];
    }
    if (((parse.commandStart+commandLength) != (script+len))
	|| ((prev == '\n') || (nested && (prev == ']')))) {
      /*
       * The command where the error occurred didn't end at the end
       * of the script (i.e. it ended at a terminator character such
       * as ";".  Reduce the length by one so that the error message
       * doesn't include the terminator character.
       */

      commandLength -= 1;
    }

    LogCompilationInfo (interp, script, parse.commandStart, commandLength);
#endif

    Tcl_FreeParse(&parse);
  }

  PTClearTokenList (tlist);
  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	PT_Traverse --
 *
 *	------------------------------------------------*
 *	This procedure traveres the tree implicitly
 *	stored in the token list, either in pre- or
 *	post-order.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Executes a callback procedure for every
 *		token enccountered.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

EXPORT (int, PT_Traverse) (interp, tokenPtr, numTokens,
			   postorder, callback, clientData)
Tcl_Interp*   interp;
Tcl_Token*    tokenPtr;
int           numTokens;
int           postorder;
PT_TokenProc* callback;
ClientData    clientData;
{
  if (postorder) {
    return PostorderTraverse (interp, tokenPtr, numTokens,
			      callback, clientData, 0);

  } else {
    return PreorderTraverse (interp, tokenPtr, numTokens,
			     callback, clientData, 0);
  }
}

/*
 *------------------------------------------------------*
 *
 *	PreorderTraverse --
 *
 *	------------------------------------------------*
 *	This procedure traveres the tree implicitly
 *	stored in the token list in preorder, i.e. it
 *	touches the root node first and then its children.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Executes a callback procedure for every
 *		token enccountered.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PreorderTraverse (interp, tokenPtr, numTokens, callback,
		  clientData, level)
Tcl_Interp*   interp;
Tcl_Token*    tokenPtr;
int           numTokens;
PT_TokenProc* callback;
ClientData    clientData;
int           level;
{
  /* Preorder traversal of the tree implied in the definition
   * of the Tcl_Token structure.
   */

  int        i, code;
  Tcl_Token* t;

  for (i=0, t = tokenPtr; i < numTokens; i++, t++) {

    code = callback (interp, clientData, level, t);

    if (code != TCL_OK) {
      /* CONTINUE, BREAK, ERROR */
      return code;
    }

    if (t->numComponents > 0) {
      code = PreorderTraverse (interp, t+1, t->numComponents,
			       callback, clientData, level + 1);
      switch (code) {
      case TCL_OK:
      case TCL_CONTINUE:
	/* Nothing to do
	 * OK       :- Go to successor, which is next on this level
	 * CONTINUE :- Skipped on level below, go to successor.
	 */
	break;

      case TCL_BREAK:
      case TCL_ERROR:
	/* Propagate outward */
	return code;
      }

      i += t->numComponents;
      t += t->numComponents;
    }
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PostorderTraverse --
 *
 *	------------------------------------------------*
 *	This procedure traveres the tree implicitly
 *	stored in the token list in postorder, i.e. it
 *	touches the children of the root node first and
 *	then the root itself.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Executes a callback procedure for every
 *		token enccountered.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PostorderTraverse (interp, tokenPtr, numTokens, callback,
		   clientData, level)
Tcl_Interp*   interp;
Tcl_Token*    tokenPtr;
int           numTokens;
PT_TokenProc* callback;
ClientData    clientData;
int           level;
{
  /* Postorder traversal of the tree implied in the definition
   * of the Tcl_Token structure.
   */

  int        i, code;
  Tcl_Token* t;

  for (i=0, t=tokenPtr; i < numTokens; i++, t++) {

    if (t->numComponents > 0) {
      code = PostorderTraverse (interp, t+1, t->numComponents,
				callback, clientData, level + 1);
      switch (code) {
      case TCL_OK:
      case TCL_CONTINUE:
	/* Nothing to do
	 * OK       :- Go to successor, which is the parent of the
	 *             nodes just processed
	 * CONTINUE :- Skipped on level below, go to successor.
	 */
	break;

      case TCL_BREAK:
      case TCL_ERROR:
	/* Propagate outward */
	return code;
      }
    }

    code = callback (interp, clientData, level, t);

    if (code != TCL_OK) {
      /* CONTINUE, BREAK, ERROR */
      return code;
    }

    if (t->numComponents > 0) {
      i += t->numComponents;
      t += t->numComponents;
    }
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	TokenTypeAsString --
 *
 *	------------------------------------------------*
 *	Converts the type of a token into a human readable
 *	string, which is then returned to the caller in
 *	the form of a String Tcl_Obj'ect.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		A new Tcl_Obj.
 *
 *------------------------------------------------------*
 */

static Tcl_Obj*
TokenTypeAsString (t)
Tcl_Token* t;
{
  char* ttype;

  switch (t->type) {
  case TCL_TOKEN_WORD:        ttype = "word";        break;
  case TCL_TOKEN_SIMPLE_WORD: ttype = "simpleWord";  break;
  case TCL_TOKEN_TEXT:        ttype = "text";        break;
  case TCL_TOKEN_BS:          ttype = "backslash";   break;
  case TCL_TOKEN_COMMAND:     ttype = "command";     break;
  case TCL_TOKEN_VARIABLE:    ttype = "var";         break;
  case TCL_TOKEN_SUB_EXPR:    ttype = "expr";        break;
  case TCL_TOKEN_OPERATOR:    ttype = "op";          break;
  case PT_TOKEN_COMMENT:      ttype = "comment";     break;
  default:
    ttype = "** unknown token type **";
  }

  return Tcl_NewStringObj (ttype, -1);
}

/*
 *------------------------------------------------------*
 *
 *	PlainList --
 *
 *	------------------------------------------------*
 *	Converts a single token into a List Tcl_Obj'ect
 *	and adds it to the List Tcl_Obj'ect provided by
 *	the 'clientData' pointer. This procedure can and
 *	will be used as callback to 'PT_Traverse' to
 *	convert a token list into a tcl list.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Extends the list object in 'clientData'.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
PlainList (interp, clientData, level, tokenPtr)
Tcl_Interp*   interp;
ClientData clientData;
int        level;
Tcl_Token* tokenPtr;
{
  int      res;
  Tcl_Obj* list = (Tcl_Obj*) clientData;
  Tcl_Obj* o    = Tcl_NewListObj (0, NULL);

  LIST_ADDOBJ  (o, TokenTypeAsString (tokenPtr));
  LIST_ADDINT  (o, level);
  LIST_ADDINT  (o, tokenPtr->numComponents);
  LIST_ADDFSTR (o, tokenPtr->start, tokenPtr->size);
  LIST_ADDOBJ  (list, o);

  return TCL_OK;

error:
  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	CallScript --
 *
 *	------------------------------------------------*
 *	Used as callback to 'PT_Traverse' to lift the
 *	C-level callback into the Tcl-level. Executes the
 *	Tcl command stored in the 'clientData' pointer
 *	for the passed token.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		As of the called Tcl command.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
CallScript (interp, clientData, level, tokenPtr)
Tcl_Interp*   interp;
ClientData clientData;
int        level;
Tcl_Token* tokenPtr;
{
  int      res;
  Tcl_Obj* cmd = (Tcl_Obj*) clientData;

  /*
   * Command called as:  cmd type level numComponents string
   */

  cmd = Tcl_DuplicateObj (cmd);
  Tcl_IncrRefCount (cmd);

  if (cmd == (Tcl_Obj*) NULL) {
    return TCL_ERROR;
  }

  LIST_ADDOBJ  (cmd, TokenTypeAsString (tokenPtr));
  LIST_ADDINT  (cmd, level);
  LIST_ADDINT  (cmd, tokenPtr->numComponents);
  LIST_ADDFSTR (cmd, tokenPtr->start, tokenPtr->size);

  /*
   * -W- Speed eval by manipulating a precompiled bytecode object.
   *
   * In future, try to manipulate an existing bytecode object
   * (compile the command once, with fake arguments, note the places
   * for the argument objects, and fill them here with the actual
   * values. Or something like that.
   *
   * Alternative: Curried commands from Feather.
   */

  res = Tcl_EvalObjEx (interp, cmd, TCL_EVAL_GLOBAL);

  Tcl_DecrRefCount (cmd);
  return res;

error:
  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	TokenObjCmd --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'parsetools::token'
 *	command. See the manpages for details on what it
 *	does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
TokenObjCmd (notUsed, interp, objc, objv)
ClientData  notUsed;		/* Not used. */
Tcl_Interp* interp;		/* Current interpreter. */
int         objc;		/* Number of arguments. */
Tcl_Obj*    objv [];		/* Arguments. */
{
  /* Syntax:  token ?-order pre|post? ?-command cmd? ?-expr 0|1? string
   *          order   = pre by default
   *          command = {} by default, causes return of plain list of tokens.
   *
   * Legal argument counts: 2, 4, 6, 8
   * Command called as:  cmd type level numComponents string
   *                         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   *                         This is the structure of the items
   *                         in the returned list in case of command = {}
   */

  char*       script;                /* The script to tokenize */
  Tcl_Obj*    cmd = (Tcl_Obj*) NULL; /* The command to call for each token */
  int         postorder = 0;
  int         stdcmd    = 1;         /* Use a standard command (PlainList) */
  int         expr      = 0;
  int         numBytes, res, i, opt, ord, len, numTokens;

  PTTokenList btlist;
  Tcl_Parse   parse;
  Tcl_Token*  tokenPtr;

  /*
   * Static information for the option processor below.
   */

  static char *option [] = {
    "-command", "-expr", "-order", NULL
  };
  enum option {
    OPT_COMMAND, OPT_EXPR, OPT_ORDER
  };	  

  static char *order [] = {
    "post", "pre", NULL
  };
  enum order {
    ORD_POST, ORD_PRE
  };	  

  /*
   * Basic check of incoming information
   */

  if ((objc < 2) || (objc > 8) || (objc % 2 == 1)) {
    Tcl_AppendResult (interp,
		      "wrong # args: should be \"", ARGSTR (0),
		      "?-order pre|post? ?-command cmd? ?-expr 0|1? string\"",
		      (char*) NULL);
    return TCL_ERROR;
  }

  /*
   * Extract the option information.
   */

  i = 1;

  while (ARGSTR(i) [0] == '-') {
    /* Process possible option.
     */

    if (Tcl_GetIndexFromObj (interp, objv [i], option, "option", 0,
			     &opt) != TCL_OK) {
      return TCL_ERROR;
    }

    switch (opt) {
    case OPT_ORDER:
      if (Tcl_GetIndexFromObj (interp, objv [i+1], order, "order", 0,
			       &ord) != TCL_OK) {
	return TCL_ERROR;
      }
      postorder = (ord == ORD_POST);
      break;

    case OPT_COMMAND:
      cmd    = objv [i+1];
      Tcl_GetStringFromObj (cmd, &len);
      stdcmd = (len == 0);
      break;

    case OPT_EXPR:
      if (Tcl_GetBoolean (interp,
			  Tcl_GetStringFromObj (objv [i+1], NULL),
			  &expr) != TCL_OK) {
	return TCL_ERROR;
      }
      break;
    }

    i += 2;
  }

  script = ARGSTRN (i, numBytes);

  /*
   * Now tokenize the string, then call the command for every generated
   * token. In case of no command given use the standard callback to
   * generate a list containing the tokens in plain form, in processing
   * order and then return that list.
   */

  if (!expr) {
    res = TokenizeWords (interp, script, numBytes, 0, &btlist);

    tokenPtr  = btlist.tokenPtr;
    numTokens = btlist.numTokens;
  } else {
    res = Tcl_ParseExpr (interp, script, numBytes, &parse);

    tokenPtr  = parse.tokenPtr;
    numTokens = parse.numTokens;
  }

  if (res == TCL_OK) {
    if (stdcmd) {
      Tcl_Obj* tlist = Tcl_NewListObj (0, NULL);
      
      res = PT_Traverse (interp, tokenPtr, numTokens,
			 postorder, PlainList, tlist);
      
      if (res != TCL_ERROR) {
	Tcl_SetObjResult (interp, tlist);
      }
    } else {
      res = PT_Traverse (interp, tokenPtr, numTokens,
			 postorder, CallScript, cmd);
      
      if (res != TCL_ERROR) {
	Tcl_SetObjResult (interp, Tcl_NewObj ());
      }
    }
  }

  if (!expr) {
    PTClearTokenList (&btlist);
  } else {
    Tcl_FreeParse (&parse);
  }

  return res;
}

/*
 *------------------------------------------------------*
 *
 *	PTInit_Token --
 *
 *	------------------------------------------------*
 *	Initialize and register token-related commands.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Tcl_CreateObjCommand.
 *
 *	Result:
 *		A standard Tcl error code.
 *
 *------------------------------------------------------*
 */

int PTInit_Token (interp)
Tcl_Interp* interp;
{
  Tcl_CreateObjCommand (interp, "parsetools::token",
			&TokenObjCmd,
			(ClientData) NULL,
			(Tcl_CmdDeleteProc*) NULL);

  return TCL_OK;
}

