# -*- tcl -*-
#		ParseTools 0.2, as of April 27, 2001
#
# CVS: $Id: ptree.tcl,v 1.7 2000/01/14 23:47:02 aku Exp $
#
# @c Handling of abstract syntax trees (AST's). ASTs are made out of
# @c nodes, both ASTs and nodes can have arbitrary information
# @c associated with them, in the form of named attributes. Each node,
# @c with the exception of the root has a single parent. Each node may
# @c have an arbitrary number of children. A node without children is
# @c a [strong leaf].
# @s Abstract syntax trees (AST's).
#
# ------------------------------------------------------------------

# ------------------------------------------------------------------
# An AST (of a script) is stored in an array A thusly:
# 
# * Each node is identified by an arbitrary key, say a number.
#
# * The key of the root node is always defined by the caller and
#   stored in A(root).
#
# * A counter is maintained for the internal generation of node keys,
#   its current value is stored in A(n) and always contains the number
#   of generated nodes. This number is greater or equal to the number
#   of existing nodes, depending on the number of 'DelNode' calls
#   issued during the construction of the tree.
#
# * The system maintains three pieces of information per node:
#
#   - A reference to the parent node of the node, stored in A(<key>,parent). 
#   - A list of children, stored in A(<key>,children).
#   - The attributes of the node. The value of each node attribute X is
#     stored in A(<key>,assoc,X). Tree attributes are stored in A(assoc,X).
#
#   The root node is an exception to the second item, it has no parent
#   node. Because of this it is not allowed to use it in calls of 'DelNode'
#   and 'DelTree'.
#
#   <key> is a placeholder for the identifying key of the node.
#
# ------------------------------------------------------------------
# The above translates in a straight forward way to a nested list
# representation and XML.
#
# In the nested list representation every node is represented by a list
# whose first element is the key of the node. The second element contains
# the associated information, the third is a list containing the
# representations of the child nodes, if any.
#
# ------------------------------------------------------------------

proc ::parsetools::ast::__Load__ {} {
    # @c Internal command. Used by other parts of the system to auto-load
    # @c the AST module.
}

# ------------------------------------------------------------------

proc ::parsetools::ast::new {tree} {
    # @c Creates a new AST with the specified name. 
    # @a tree:     The handle/name of the new AST.
    # @r The object command for the tree.

    upvar ::parsetools::ast::$tree t

    array set t {
	seq      0
	size     1
	0,parent	{}
	0,children	{}
    }

    interp alias {} $tree      {} ::parsetools::ast::TreeCmd $tree
    interp alias {} ${tree}.n0 {} ::parsetools::ast::NodeCmd $tree 0
    return $tree
}


proc ::parsetools::ast::TreeCmd {tree method args} {
    # @c Internal command. Implementation of the tree object commands.
    # @c Called via aliasing.
    # @a tree:     The handle/name of the AST.
    # @a method:   The name of the called method.
    # @a args:     Additional arguments for the method.

    eval T$method [list $tree] $args
}


proc ::parsetools::ast::NodeCmd {tree node method args} {
    # @c Internal command. Implementation of the node object commands.
    # @c Called via aliasing.
    # @a tree:     The handle/name of the AST.
    # @a node:     The internal id of the node.
    # @a method:   The name of the called method.
    # @a args:     Additional arguments for the method.

    eval N$method [list $tree] [list $node] $args
}

# ------------------------------------------------------------------

proc ::parsetools::ast::Tdelete {tree} {
    # @c Deletes the entire AST.
    # @a tree: The handle/name of the AST to delete.

    # Remove all nodes and their object commands, then clear the
    # remaining array and the root.

    foreach n [[$tree root] children] {
	$n delete
    }

    unset ::parsetools::ast::$tree
    interp alias {} $tree      {}
    interp alias {} ${tree}.n0 {}
    return
}


proc ::parsetools::ast::Tsize {tree} {
    # @c Retrieves the number of nodes in the AST.
    # @a tree: The handle/name of the AST to query.

    upvar ::parsetools::ast::$tree t
    return   $t(size)
}


proc ::parsetools::ast::Troot {tree} {
    # @c Retrieves the object command of the root node in the AST.
    # @a tree:     The handle of the AST.

    return   ${tree}.n0
}


proc ::parsetools::ast::Ttraverse {tree command} {
    # @c Traversing a AST. Calls the <a command> for the tree, once
    # @c before and once after traversing all nodes, and for all nodes in
    # @c the tree. Does a pre-order traversal, i.e. touches the root of the
    # @c tree first and its children afterward. The <a command> is always
    # @c executed in the calling context and thus has access to all variables
    # @c there. This eespecially allows it to store the state of the traversal
    # @c in the variables of a procedure surrounding the call to the traverser.
    #
    # @a tree:     The handle/name of the AST to traverse.
    # @a command:  The command called for the tree and each node.

    uplevel 2 $command tree $tree enter

    _Traverse $tree 0 3 0 $command

    uplevel 2 $command tree $tree leave
}


proc ::parsetools::ast::_Traverse {tree level uplevel node command} {
    # @c Internal command for traversing the AST. Calls the
    # @c <a command> for the current <a node>, once before and once
    # @c after traversing the childs of the current node.
    # @a tree:     The handle/name of the AST to traverse.
    # @a level:    How deep we are in the tree.
    # @a uplevel:  The number of level to find the context calling the
    # @a uplevel:  traverser.
    # @a node:     The internal id of the current node.
    # @a command:  The command called for each node (twice).

    upvar ::parsetools::ast::$tree t

    uplevel $uplevel $command node ${tree}.n$node enter $level
    incr level
    incr uplevel

    foreach subnode $t($node,children) {
	_Traverse $tree $level $uplevel $subnode $command
    }

    incr uplevel -1
    incr level -1
    uplevel $uplevel $command node ${tree}.n$node leave $level
    return
}


proc ::parsetools::ast::Tset {tree attr value} {
    # @c Associates the AST with arbitrary information (attributes).
    # @a tree:     The handle/name of the AST.
    # @a attr:     The name of the attribute.
    # @a value:    The value of the attribute.
    # @r the <a value>, of an arbitrary type

    upvar ::parsetools::ast::$tree t
    set t(assoc,$attr) $value
}


proc ::parsetools::ast::Tget {tree {attr {}}} {
    # @c Queries the AST for its attributes and their values.
    # @c If an attribute name was specified the value of that
    # @c attribute is returned. Else a list containing all
    # @c attribute names and their values is returned, in a
    # @c form accepted by [strong {array set}].
    #
    # @a tree:   The handle/name of the AST.
    # @a attr:   The name of the queried attribute, possibly empty.
    # @r the requested value, of arbitrary type; or a list.

    upvar ::parsetools::ast::$tree t

    if {$attr == {}} {
	set res [list]
	foreach item [array names t assoc,*] {
	    lappend res [lindex [split $item ,] 1] $t($item)
	}
	return $res
    } else {
	return $t(assoc,$attr)
    }
}

proc ::parsetools::ast::Tdel {tree attr} {
    # @c Removes an attribute from the tree.
    # @a tree:    The handle/name of the AST.
    # @a attr:    The name of the attribute to remove.

    upvar ::parsetools::ast::$tree t
    catch {unset t(assoc,$attr)}
}


# ------------------------------------------------------------------

proc ::parsetools::ast::Nset {tree node args} {
    # @c Associates the node with arbitrary information (attributes).
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query.
    # @a attr:    The name of the attribute.
    # @a value:   The value of the attribute.
    # @r the <a value>, of arbitrary type.

    upvar ::parsetools::ast::$tree t

    if {[llength $args] %2 == 1} {
	error "Incorrect number of arguments. Must be even."
    }

    foreach {attr value} $args {
	set t($node,assoc,$attr) $value
    }
}


proc ::parsetools::ast::Ndel {tree node attr} {
    # @c Removes an attribute from the node.
    # @a tree:   The handle/name of the AST.
    # @a node:   The internal id of the node to query.
    # @a attr:   The name of the attribute to remove.

    upvar ::parsetools::ast::$tree t
    catch {unset t($node,assoc,$attr)}
}


proc ::parsetools::ast::Nget {tree node {attr {}}} {
    # @c Queries the node  for its attributes and their values.
    # @c If an attribute name was specified the value of that
    # @c attribute is returned. Else a list containing all
    # @c attribute names and their values is returned, in a
    # @c form accepted by [strong {array set}].
    # @a tree:   The handle/name of the AST.
    # @a node:   The internal id of the node to query.
    # @a attr:   The name of thee queried attribute, possibly empty.
    # @r the requested value, of arbitrary type; or a list.

    upvar ::parsetools::ast::$tree t

    if {$attr == {}} {
	set res [list]
	foreach item [lsort [array names t $node,assoc,*]] {
	    lappend res [lindex [split $item ,] 2] $t($item)
	}
	return $res
    } else {
	return $t($node,assoc,$attr)
    }
}


proc ::parsetools::ast::Nparent {tree node} {
    # @c Asks a node for its parent.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query.
    # @r the object command of the parent to the queried node.

    upvar ::parsetools::ast::$tree t
    return   $tree.n$t($node,parent)
}


proc ::parsetools::ast::NisLeaf {tree node} {
    # @c Asks a node wether it is a leaf (= without children) or not.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query.
    # @r a boolean value.

    upvar ::parsetools::ast::$tree t
    expr   {[llength $t($node,children)] == 0}
}


proc ::parsetools::ast::N#children {tree node} {
    # @c Asks a node about the number of its children.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query.
    # @r a number.

    upvar ::parsetools::ast::$tree t
    llength $t($node,children)
}


proc ::parsetools::ast::Nchildren {tree node {from {}} {to {}}} {
    # @c Asks a node for the child nodes it contains
    # @a tree: The handle/name of the AST.
    # @a node: The internal id of the node to query.
    # @a from: Optional argument. Defines together with <a to> the range in
    # @a from: the list of children the caller asks for. Defaults to 1.
    # @a to:   Optional argument. Defines together with <a to> the range in
    # @a to:   the list of children the caller asks for. Defaults to 'end'.
    #
    # @r a list containing the object commands of the nodes which are the
    # @r children of the queried node, in the specified range.

    upvar ::parsetools::ast::$tree t

    if {[string equal $from {}]} {
	# Neither 'from' nor 'to' were set => assume 1..end
	set children $t($node,children)
    } elseif {[string equal $to {}]} {
	# 'from' was set, but not 'to' => assume 'end'
	set children [lrange $t($node,children) $from end]
    } else {
	set children [lrange $t($node,children) $from $to]
    }

    set res [list]

    foreach n $children {
	lappend res ${tree}.n$n
    }

    return $res
}


proc ::parsetools::ast::Nchild {tree node idx} {
    # @c Asks a node for the child nodes it contains
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query.
    # @a idx:     The index of the child to retrieve.
    # @r the object command of the child we were asked for.

    upvar ::parsetools::ast::$tree t

    return ${tree}.n[lindex $t($node,children) $idx]
}


proc ::parsetools::ast::Ntree {tree node} {
    # @c Asks a node for the AST it belongs to.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query.
    # @r the object command of a AST

    return $tree
}


proc ::parsetools::ast::NaddChild {tree node} {
    # @c Adds a new node to the AST, with the
    # @c specified <a node> as its parent.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to query
    # @a node:    the new node is added to.
    # @r the object command of the new node.

    upvar ::parsetools::ast::$tree t

    set key [incr t(seq)]
    array set t [list $key,parent $node $key,children {}]

    lappend t($node,children) $key
    incr    t(size)

    #puts "NaddChild $tree $node => $key"

    interp alias {} ${tree}.n$key {} ::parsetools::ast::NodeCmd $tree $key
    return  ${tree}.n$key
}


proc ::parsetools::ast::Ndelete {tree node} {
    # @c Deletes the specified <a node> and its children.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to delete.

    upvar ::parsetools::ast::$tree t

    #puts "Ndelete $tree $node"

    foreach subnode $t($node,children) {
	Ndelete $tree $subnode
    }

    Ncut $tree $node
    return
}


proc ::parsetools::ast::Nsplice {tree node from to} {
    # @c Takes the children of <a node> in the range <a from>..<a to> and
    # @c replaces them with a new node. The specified children get the new
    # @c node as their parent. This command is the complement to
    # @c <p ::parsetools::ast::Ncut>.
    # @a tree:  The handle/name of the AST.
    # @a node:  The internal id of the node to query.
    # @a from:  The index of the first child to move down.
    # @a to:    The index of the last  child to move down.

    upvar ::parsetools::ast::$tree t

    set spliced [lrange $t($node,children) $from $to]

    set key [incr t(seq)]
    array set t [list $key,parent $node $key,children $spliced]

    set t($node,children) [lreplace $t($node,children) $from $to $key]

    foreach subnode $spliced {
	set t($subnode,parent) $key
    }

    interp alias {} ${tree}.n$key {} ::parsetools::ast::NodeCmd $tree $key
    return  ${tree}.n$key
}


proc ::parsetools::ast::Ncut {tree node} {
    # @c Deletes the specified <a node>, but [strong not] its children.
    # @c The children are retained and made into children of the parent
    # @c of the removed node. Order is preserved, i.e. the new children
    # @c of our parent occupy the same place as the removed itself.
    # @a tree:    The handle/name of the AST.
    # @a node:    The internal id of the node to remove.

    # not recursive !
    upvar ::parsetools::ast::$tree t

    set p $t($node,parent)

    #puts "Ncut $tree $node"

    set nref [lsearch $t($p,children) $node]

    if {[llength $t($node,children)] > 0} {
	# Lift our children into the parent, in
	# the place of the reference to ourselves.

	set t($p,children) \
		[eval lreplace [list $t($p,children)] \
		$nref $nref $t($node,children)]

	# Go through the list of children and correct their
	# parent pointer too.

	foreach c $t($node,children) {
	    set t($c,parent) $p
	}

    } else {
	# Just remove the reference to us from our parent.

	set t($p,children) [lreplace $t($p,children) $nref $nref]
    }

    foreach key [array names t $node,*] {
	unset t($key)
    }

    incr t(size) -1
    return
}

