# -*- tcl -*-
#		ParseTools 0.2, as of April 27, 2001
#
# CVS: $Id: cmd.tcl,v 1.2 2000/01/14 23:47:02 aku Exp $
#
# @c A database allowing the association of arbitrary information with
# @c commands (-> command attributes). Future vision is to have that
# @c database built into the core interpreter and new commands (from
# @c extension and packages) automatically filling it with some standard
# @c information, f.e. about arguments, return values, semantic promises
# @c (like: 'does not affect/fire traces'), etc.
# @s Database for commands and associated attributes.
#
# ------------------------------------------------------------------

namespace eval ::parsetools::cmd {
    # This namespace will contain the procedures and variables
    # managing a database of command and associated attributes.

    variable  db
    array set db {}
}


proc ::parsetools::cmd::__Load__ {} {
    # @c Internal command. Used by other parts of the system to enforce
    # @c the loading of this module.
}


proc ::parsetools::cmd::remove {name} {
    # @c Removes all information about the named command from the database.
    # @a name: The name of the command whose information shall be removed.

    variable db

    foreach key [array names $name,*] {
	unset db($key)
    }

    return
}


proc ::parsetools::cmd::set {name attr value} {
    # @c Assign a value to an attribute of the named command.
    # @c Defines a new attribute if it does not exist already.
    # @a name: The name of the command whose attribute is manipulated.
    # @a attr: The name of the attribute.
    # @a value: The new value of the attribute.
    # @r the <a value>.

    variable db
    ::set    db($name,$attr) $value
}


proc ::parsetools::cmd::lappend {name attr value} {
    # @c Manipulates an existing attribute of the named command.
    # @c The specified value is appended to the current value of
    # @c the attribute, while treating that one as a list.
    # @a name: The name of the command whose attribute is manipulated.
    # @a attr: The name of the attribute.
    # @a value: The data to add to the attribute.
    # @r the new value of the attribute.

    variable db
    lappend  db($name,$attr) $value
    ::set    db($name,$attr)
}


proc ::parsetools::cmd::append {name attr value} {
    # @c Manipulates an existing attribute of the named command.
    # @c The specified value is appended to the current value of
    # @c the attribute, while treating that one as a string.
    # @a name: The name of the command whose attribute is manipulated.
    # @a attr: The name of the attribute.
    # @a value: The data to add to the attribute.
    # @r the new value of the attribute.

    variable db
    append   db($name,$attr) $value
    ::set    db($name,$attr)
}


proc ::parsetools::cmd::get {name attr} {
    # @c Retrieves the current value of the specified attribute for
    # @c the named command.
    # @a name: The name of the command whose attribute is queried
    # @a attr: The name of the queried attribute.

    variable db
    ::set    db($name,$attr)
}


proc ::parsetools::cmd::getattrs {name} {
    # @ Retrieves the current values of all attributes for the named command.
    # @a name: The name of the queried command.
    # @r a list containing the names of the attributes associated
    # @r to the specified command and their values, in a form
    # @r accepted by [strong {array set}].

    variable db

    ::set res [list]
    foreach key [array names $name,*] {
	lappend res [lindex [split $key ,] 1] $db($key)
    }

    ::set res
}


proc ::parsetools::cmd::del {name attr} {
    # @c Deletes the specified attribute for the named command.
    # @a name: The name of the command whose attribute is manipulated.
    # @a attr: The name of the attribute to delete.

    variable db
    unset    db($name,$attr)
    return
}

# ------------------------------------------------------------------

