# -*- tcl -*-
#		ParseTools 0.2, as of April 27, 2001
#
# CVS: $Id: convert.tcl,v 1.4 1999/12/11 17:30:34 aku Exp $
#
# @c Conversion of the syntax tree into several other representations.
# @s syntax tree conversions.

::parsetools::ast::__Load__

# ------------------------------------------------------------------

proc ::parsetools::ast::RepStr {n text} {
    # @c Replicates <a n> times the string <a text>. Originally from
    # @c <xref Pool>, copied here to avoid the dependency.
    #
    # @r The <a text>, replicated <a n> times.
    #
    # @a text: string to replicate.
    # @a n:    Number of replications. A value < 0 is handled as 0
    #
    # @i replication
    #
    # @note This algorithm proved to be rather fast, even in the
    # @note small range of n in {60..90}. An alternative definition
    # @note using regsub and format with O(n) complexity, thus doing
    # @note most things at C-level is beaten regularly for n > 74!

    if {$n <= 0} {
	# No replication required

	return ""
    } elseif {$n == 1} {
	# Quick exit for recursion

	return $text
    } elseif {$n == 2} {
	# Another quick exit for recursion

	return $text$text
    } elseif {0 == ($n % 2)} {
	# Halving the problem results in O (log n) complexity.

	set result [RepStr [expr {$n / 2}] $text]
	return "$result$result"
    }

    # Uneven length, reduce problem by one
    return "$text[RepStr [incr n -1] $text]"
}


# ------------------------------------------------------------------
# Generic conversion into nested list

proc ::parsetools::ast::TtoNestedList {tree} {
    # @c Converts the syntax tree in question into a nested listed
    # @c representation.
    # @a tree: The handle/name of the tree to manipulate.
    # @r a nested listed

    variable stack [list]

    Ttraverse $tree _toNestedList

    set     res [lindex $stack 0]
    unset                stack

    list [$tree get] $res
}


proc ::parsetools::ast::_toNestedList {type obj op args} {
    # @c See <p ::parsetools::ast::TtoNestedList>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    variable stack

    if {[string equal $op enter]} {
	# Ignore enter, not relevant here.
	return
    }
    if {[string equal $type tree]} {
	# Ignore the tree calls, not relevant here.
	return
    }

    # Leaving a node here now.

    # Has n children => Take that many items from the stack,
    # combine with the current node and place the result back.

    set  numc [$obj #children]

    if {$numc == 0} {
	set            node [list]
	lappend        node $obj [$obj get] {}
	lappend stack $node
    } else {
	incr numc -1

	set           node [list]
	lappend       node $obj [$obj get] [lrange $stack end-$numc end]
	set stack     [lreplace $stack end-$numc end $node]
    }
    return
}

# ------------------------------------------------------------------

proc ::parsetools::ast::TtoTypedNestedList {tree} {
    # @c Converts the syntax tree in question into a nested listed
    # @c representation using the associated 'type' as main element
    # @c of each node.
    # @a tree: The handle/name of the tree to manipulate.
    # @r a nested listed

    variable stack [list]

    Ttraverse $tree _toTypedNL

    set     res [lindex $stack 0]
    unset                stack
    return $res
}


proc ::parsetools::ast::_toTypedNL {type obj op args} {
    # @c See <p ::parsetools::ast::TtoTypedNestedList>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    variable stack

    if {[string equal $op enter]} {
	# Ignore enter, not relevant here.
	return
    }
    if {[string equal $type tree]} {
	# Ignore the tree calls, not relevant here.
	return
    }

    # Leaving a node here now.

    # Has n children => Take that many items from the stack,
    # combine with the current node and place the result back.

    set  numc [$obj #children]

    array set x [$obj get]
    set type $x(type)
    unset     x(type)

    if {$numc == 0} {
	set            node [list]
	lappend        node $type [array get x] {}
	lappend stack $node
    } else {
	incr numc -1

	set           node [list]
	lappend       node $type [array get x] [lrange $stack end-$numc end]
	set stack     [lreplace $stack end-$numc end $node]
    }
    return
}

# ------------------------------------------------------------------
# Generic conversion into XML

proc ::parsetools::ast::TtoXML {tree} {
    # @c Converts the syntax tree in question into XML.
    # @a tree: The handle/name of the tree to manipulate.
    # @r a string containing XML structures.

    variable txt ""
    variable level

    Ttraverse $tree _toXML

    set     res $txt
    unset        txt
    return $res
}


proc ::parsetools::ast::_toXML {type obj op args} {
    # @c See <p ::parsetools::ast::TtoXML>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    variable txt

    switch -- $type,$op {
	tree,enter {
	    append txt "<tree name=\"$obj\">\n"
	}
	tree,leave {
	    append txt "</tree>"
	}
	node,enter {
	    set  level [lindex $args 0]
	    incr level

	    append txt "[RepStr $level "    "]<node name=\"$obj\">\n"

	    set data [$obj get]

	    if {[llength $data] > 0} {
		incr level

		foreach {key value} $data {
		    append txt [RepStr $level "    "]
		    append txt "<item key=\"$key\">"
		    append txt $value
		    append txt </item>\n
		}
	    }
	}
	node,leave {
	    set  level [lindex $args 0]
	    incr level

	    append txt "[RepStr $level "    "]</node>\n"
	}
    }
}

# ------------------------------------------------------------------
# Second generic conversion into XML (types as node tags)

proc ::parsetools::ast::TtoTypedXML {tree} {
    # @c Converts the syntax tree in question into XML using the
    # @c associated 'type' as tag name for each node.
    # @a tree: The handle/name of the tree to manipulate.
    # @r a string containing XML structures.

    variable txt ""
    variable level

    Ttraverse $tree _toTypedXML

    set     res $txt
    unset        txt
    return $res
}


proc ::parsetools::ast::_toTypedXML {type obj op args} {
    # @c See <p ::parsetools::ast::TtoTypedXML>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    variable txt

    switch -- $type,$op {
	tree,enter {
	    append txt "<tree name=\"$obj\">\n"
	}
	tree,leave {
	    append txt "</tree>"
	}
	node,enter {
	    set level [lindex $args 0]
	    set type  [$obj get type]
	    incr level

	    append txt "[RepStr $level "    "]<$type>\n"

	    set data [$obj get]

	    if {[llength $data] > 2} {
		incr level

		foreach {key value} $data {
		    if {[string equal $key type]} {
			# Skip type information, is in node tag.
			continue
		    }

		    append txt [RepStr $level "    "]
		    append txt "<item key=\"$key\">"
		    append txt $value
		    append txt </item>\n
		}
	    }
	}
	node,leave {
	    set level [lindex $args 0]
	    set type  [$obj get type]
	    incr level

	    append txt "[RepStr $level "    "]</$type>\n"
	}
    }
}

# ------------------------------------------------------------------

