# -*- tcl -*-
#		ParseTools 0.2, as of April 27, 2001
#
# CVS: $Id: parse.tcl,v 1.7 1999/12/11 20:57:13 aku Exp $
#
# @c High level parsing functionality using the tcl-level tokenizer
# @s Higher level parsing functionality

::parsetools::ast::__Load__

# ------------------------------------------------------------------
# -W-, future: Use a database of command descriptions containing more
# -W-, future: information about the words of a command, f.e. which words
# -W-, future: contain expressions or scripts (if, while, ...). Use this
# -W-, future: to parse the script deeper than usual, i.e. deeper than done
# -W-, future: by the real tcl parser.

# -W-, future: translate array, nested list, XML back into a script.

# -W-, future: XML -> syntax tree -> token list -> script conversions
# -W-, future: disassembler, assembler commands for bytecodes ?
# -W-, future: application: static code checker
# -W-, future: application: use as parse engine for autodoc.
# -W-, future: application: cross referencer (proc/var declare <-> usage, ...)
# -W-, future: application: autodoc + cross ref. -> single app.

# -W-, future: If the first word of the command is a constant text (stext)
# -W-, future: we can use the above mentioned database of command
# -W-, future: descriptions to check out the specialities of the argument
# -W-, future: words. If the special words are constant too we can recurse
# -W-, future: and go deeper than the regular parser.
# ------------------------------------------------------------------

# ------------------------------------------------------------------
# Standard conversion: Token list into syntax tree, no changes to the
# structure.

proc ::parsetools::tokensToTree {tree tokens} {
    # @c Convert a list of tokens into an equivalent syntax tree. No changes
    # @c are made to the tree.
    # @a tree:   The name of the new tree, the name of the object command to
    # @a tree:   access the new tree. An existing tree of that name is deleted.
    # @a tokens: The list of tokens to convert.
    # @r <a tree>.

    catch {$tree delete}
    ast::new $tree

    TTT [$tree root] $tokens
    [$tree root] set type script
    return $tree
}


proc ::parsetools::TTT {node tokens} {
    # @c Internal command. Used by <p ::parsetools::tokensToTree> to do the
    # @c main work. Recurses down into the implicit tree while building it
    # @c explicitly.
    # @a node: The handle / object command of the node the new found children
    # @a node: are added to.
    # @a tokens: The list of tokens to convert.

    while {[llength $tokens] > 0} {
	# Take the first token in the list, its children and add all of them
	# to the tree. Repeat until exhaustion of the list of tokens.

	set firstToken [lindex $tokens 0]

	foreach {type level n data} $firstToken break

	set child [$node addChild]
	$child set type $type
	$child set text $data

	TTT $child [lrange $tokens 1 $n]
	set tokens [lrange $tokens [incr n] end]
    }
}

# ------------------------------------------------------------------
# Convert script into a standard tree.

proc ::parsetools::scriptToTree {tree script} {
    # @c Parses the script into a list of tokens and converts these then
    # @c into the equivalent standard syntax tree.
    # @a tree:   The name of the new tree, the name of the object command to
    # @a tree:   access the new tree. An existing tree of that name is deleted.
    # @a script:   The script to parse and store in the array.
    # @r <a tree>.

    # Node types in the standard tree:
    # word       - A word made up out of parts like constant strings,
    #              bracket-commands and variable references.
    # simpleWord - A word made out of a constant string (single child)
    # text       - A constant string.
    # backslash  - A constant string, in the form of a backslash sequence.
    # var        - Variable reference: First child names it, optional second
    #              child defines index.
    # command    - immediately below the root a simple command. As part of a
    #              word a bracketed command [] to evaluate while creating the
    #              string in the word.
    # comment    - A comment
    # expr       - Expression string, first child is operator, followed by
    #              the subexpressions.
    # op         - Operator definition in eexpression.

    tokensToTree $tree [token $script]
}

# ------------------------------------------------------------------

proc ::parsetools::compactTree {tree} {
    # @c Takes a standard syntax tree as generated by
    # @c <p ::parsetools::scriptToTree> and removes redundant information.
    # @c Afterward the tree contains less attributes and its size is reduced
    # @c as much as possible without losing information.
    # @a tree: The handle/name of the tree to manipulate.

    # Nodes in a compact tree: Same as in standard tree, but without
    # 'simpleWord'. 'text' is enough for that.

    $tree traverse ::parsetools::_compactTree
}

proc ::parsetools::_compactTree {type obj op args} {
    # @c See <p ::parsetools::compactTree>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    switch -- $type,$op {
	tree,enter {}
	tree,leave {}
	node,enter {
	    set type [$obj get type]

	    if {[string equal text      $type]} {return}
	    if {[string equal backslash $type]} {return}
	    if {[string equal comment   $type]} {return}
	    if {[string equal command   $type] && [$obj isLeaf]} {return}

	    # Remove redundant fragments of the script,
	    # leave only the necessary constants.

	    $obj del text
	}
	node,leave {
	    # We are doing this while leaving nodes so that the children are
	    # minimal before we embark to change their parent. The tree
	    # contracts from the bottom to the top.

	    set type [$obj get type]

	    if {
		[string equal simpleWord $type] ||
		([string equal word $type] && ([$obj #children] == 1))
	    } {
		# Remove all 'simpleWords' and all 'words' with a single child.
		$obj cut
	    }
	}
    }
}

# ------------------------------------------------------------------

proc ::parsetools::expandBrackets {tree} {
    # @c Takes a standard or compacted syntax tree as generated by
    # @c <p ::parsetools::scriptToTree> and expands any internal bracketed
    # @c commands by parsing them too.
    # @a tree: The name of the tree to manipulate.

    $tree traverse ::parsetools::_expandBrackets
}

proc ::parsetools::_expandBrackets {type obj op args} {
    # @c See <p ::parsetools::expandBrackets>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    switch -- $type,$op {
	tree,enter {}
	tree,leave {}
	node,enter {
	    set type [$obj get type]

	    if {!([string equal command $type] && [$obj isLeaf])} {return}

	    # Uses the internal command of the token -> tree converter to
	    # create a new subtree.

	    # We are doing this upon entry to allow the traversal to
	    # immediately go down into the new nodes, possibly expanding
	    # more newly revealed commands.

	    set text [string range [$obj get text] 1 end-1]
	    $obj                         del text

	    TTT $obj [token $text]

	    if {[$obj #children] == 1} {
		# Remove the first and only child of the expanded command,
		# as it is a redundant command node.

		[lindex [$obj children] 0] cut
	    } else {
		# The brackets contained more than one command. Retain all
		# children and change this node to be a script.

		$obj set type script
	    }
	}
	node,leave {}
    }
}

# ------------------------------------------------------------------

proc ::parsetools::varStructure {tree} {
    # @c Takes a standard syntax tree and adds more structure into variable
    # @c references (better marking of base name and index parts).
    # @a tree: The handle/name of the tree to manipulate.

    $tree traverse ::parsetools::_varStructure
}

proc ::parsetools::_varStructure {type obj op args} {
    # @c See <p ::parsetools::varStructure>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    switch -- $type,$op {
	tree,enter {}
	tree,leave {}
	node,enter {}
	node,leave {
	    # Ignore everything except 'var' references.

	    set type [$obj get type]

	    if {![string equal var $type]} {return}

	    [$obj splice 0 0] set type vbase

	    if {[$obj #children] > 1} {
		[$obj splice 1 end] set type vindex
	    }
	}
    }
}


# ------------------------------------------------------------------------

proc ::parsetools::treeToTokens {tree} {
    # @c Traverses the specified <a tree> and generates the original token
    # @c form along the way. Uses the attribute 'tokens' to associate
    # @c intermediate results with the nodes in the tree thus obviating the
    # @c need for separate state variables containing the state of the
    # @c processor.
    # @a tree: The handle/name of the tree we are working on.
    # @r a list of tokens.

    $tree traverse ::parsetools::_treeToTokens

    set tokens [$tree get tokens]
    $tree             del tokens

    return $tokens
}

proc ::parsetools::_treeToTokens {type obj op args} {
    # @c See <p ::parsetools::treeToTokens>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    switch -- $type,$op {
	tree,enter {}
	tree,leave {
	    $tree set tokens [[$tree root] get tokens]
	    [$tree root] del tokens
	}
	node,enter {}
	node,leave {
	    if {[$obj isLeaf]} {
		$obj set tokens \
			[list [list [$obj get type] 0 0 [$obj get text]]]
	    } else {
		set childtokens [list]
		foreach child [$obj children] {
		    eval lappend childtokens [$obj get tokens]
		    $obj del tokens
		}
		set mytoken [list [$obj get type] 0 [llength $childtokens] {}]
		eval lappend mytoken $childtokens

		$obj set tokens $mytoken
	    }
	}
    }
}

# ------------------------------------------------------------------------

proc ::parsetools::treeToScript {tree} {
    # @c Traverses the specified <a tree> and generates the original script
    # @c form along the way. Uses the attribute 'scriptvalue' to associate
    # @c intermediate results with the nodes in the tree thus obviating the
    # @c need for separate state variables containing the state of the
    # @c processor.
    # @a tree: The handle/name of the tree we are working on.
    # @r a tcl script

    $tree traverse ::parsetools::_treeToScript

    set script [$tree get scriptvalue]
    $tree             del scriptvalue

    # Remove outer pair of brackets.

    return [string range $script 1 end-1]
}

proc ::parsetools::_treeToScript {type obj op args} {
    # @c See <p ::parsetools::treeToScript>.
    # @a type: The type of the <a obj>ect we are processing, either 'tree'
    # @a type: or 'node'.
    # @a obj:  The handle of the current object.
    # @a op:   Either 'enter' or 'leave', allowing us to distinguish between
    # @a op:   the two calls beforee and after processing the children of
    # @a op:   the <a obj>ect in question.
    # @a args: Additional arguments. For nodes the first of them contains
    # @a args: the nesting level of node in the tree.

    switch -- $type,$op {
	tree,enter {}
	tree,leave {
	    $obj set scriptvalue [[$obj root] get scriptvalue]
	    [$obj root] del scriptvalue
	}
	node,enter {}
	node,leave {
	    set level [lindex $args 0]

	    switch -- [$obj get type] {
		script      {
		    # -W- use level for proper indentation.

		    # The children are commands or comments. Take their
		    # script values, remove the brackets, place them on
		    # separate lines. Add brackets surounding the whole
		    # script.

		    set sv ""
		    foreach c [$obj children] {
			set svc [$c get scriptvalue]

			if {[string equal command [$c get type]]} {
			    append sv "[string range $svc 1 end-1]\n"
			} else {
			    # Contains the necessary \n
			    append sv $svc
			}

			$c del scriptvalue
		    }
		    #set sv [string trim $sv]
		    $obj set scriptvalue \[$sv\]
		}
		command     {
		    # Take the script values of all children, concatenate
		    # them using space as separator, add surounding brackets.

		    set sv ""
		    foreach c [$obj children] {
			switch -- [$c get type] {
			    text -
			    simpleWord {
				append sv "[list [$c get scriptvalue]] "
			    }
			    default {
				append sv "[$c get scriptvalue] "
			    }
			}

			$c del scriptvalue
		    }
		    set sv [string trim $sv]
		    $obj set scriptvalue \[$sv\]
		}

		vindex      {
		    # Like 'word' below, but without the enclosing apostrophes.

		    set sv ""
		    foreach c [$obj children] {
			append sv [$c get scriptvalue]
			$c             del scriptvalue
		    }
		    $obj set scriptvalue $sv
		}
		word        {
		    # Take the script values of all children, concatenate
		    # them without separators. Enclose the result into
		    # apostrophes.

		    set sv ""
		    foreach c [$obj children] {
			append sv [$c get scriptvalue]
			$c             del scriptvalue
		    }
		    $obj set scriptvalue \"$sv\"
		}

		vbase       -
		simpleWord  {
		    # Move script value of single child one up.

		    set c [$obj child 0]

		    $obj set scriptvalue [$c get scriptvalue]
		    $c                       del scriptvalue
		}

		expr        {error "Handling unknown"}
		op          {error "Handling unknown"}

		var         {
		    # Several ways of having a variable reference:
		    # 1 child,    a vbase
		    # 1 child,    whatever (except vbase)
		    # 2 children, a vbase and a vindex
		    # n children, whatever (except vbase, vindex),
		    #             first is base, tail is index

		    switch -- [$obj #children] {
			1 {
			    set c [$obj child 0]
			    $obj set scriptvalue \$[$c get scriptvalue]
			    $c                         del scriptvalue
			}
			2 {
			    set base [$obj child 0]
			    set idx  [$obj child 1]

			    set    vref "\$"
			    append vref  [$base get scriptvalue]
			    append vref ([$idx get  scriptvalue])

			    $obj set scriptvalue $vref

			    $base del scriptvalue
			    $idx  del scriptvalue
			}
			default {
			    # First is basename, tail defines index
			    # (implicit node 'word').

			    set base [$obj child 0]
			    set idx  

			    set    vref "\$[$base get scriptvalue]("
			    $base                 del scriptvalue

			    foreach c [$obj children 1 end] {
				append vref [$c get scriptvalue]
				$c              del scriptvalue
			    }

			    append vref )

			    $obj set scriptvalue $vref
			}
		    }
		}

		comment     -
		text        -
		backslash   {
		    # Constant strings, copy unchanged

		    $obj set scriptvalue [$obj get text]
		}

		defered {
		    # Move script value of single child (command or script) one
		    # up, transform outer brackets into braces along the way.

		    # -W- preserve indentation.

		    set c [$obj child 0]

		    $obj set scriptvalue \
			    \{[string range [$c get scriptvalue] 1 end-1]\}
		    $c   del scriptvalue
		}

		default {
		    error "Unknown node type $type, unable to generate script"
		}
	    }
	}
    }
}

# ------------------------------------------------------------------------

